/*-----------------------------------------------------------------
  nokia_clock by Jack Botner

  main.c

  This project consists of the following hardware:

  - Nokia 5110 LCD display 48 x 84 pixels
  - pushbutton rotary encoder model EC11-1B-18T
    (Changzhou Xinze Electronic Co.) 

  A 32.768 KHz watch crystal is connected between TOSC1 and
  TOSC2 to provide timing for the clock.

  Processor: ATMega88PA
      Clock: Internal, 4 MHz (250 ns)

  Released under the MIT License  - see
  http://www.opensource.org/licenses/mit-license.php

  20120427 - program written.
-----------------------------------------------------------------*/
#include <avr/io.h>
#include <avr/sleep.h>
#include <avr/interrupt.h>
#include <util/delay_basic.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "main.h"
#include "n3310_lcd.h"

#define STATE_INITIAL		1
#define STATE_PROGRAM		2
#define STATE_RUN			3
#define STATE_SLEEP			4

// Global variables

static volatile uint8_t	ucSystemState=STATE_INITIAL;

volatile unsigned char	ucHours=12;
volatile unsigned char	ucMinutes=0;
volatile unsigned char	ucSeconds=0;
volatile unsigned char	ucFormat=FORMAT_12HRS;

volatile uint8_t  ucEventFlags=0;			// system event flags
volatile uint8_t  ucDebouncedState=0;    	// debounced state of the switches
volatile uint16_t uiMsPressed=0;			// how long encoder button pressed
volatile uint8_t  ucStopClock=0;			// to stop the clock during programming

// Internal functions

static void display_time(void);
static void system_init(void);

/*-----------------------------------------------------------------
  main()
-----------------------------------------------------------------*/
int main( void )
{
  while (1)
  {
    switch ( ucSystemState )
    {
      case STATE_INITIAL:
	    system_init();                // initialize system
	    LCD_init();					  // initialize LCD
	    ucSystemState = STATE_RUN;
		set_bit( PORTC, PC5 );		  // diagnostic led on
		
		// What if we power up with only battery power? We
		// need to come up in the right state.
		if ( !test_bit( PIND, PD2 ) )
          set_bit( ucEventFlags, EVENT_POWER_DOWN );
	    break;

      case STATE_RUN:
        if ( test_bit( ucEventFlags, EVENT_BUTTON_PRESSED ) )
        {
	      ucSystemState = STATE_PROGRAM;
	      clear_bit( ucEventFlags, EVENT_BUTTON_PRESSED );
        }

        if ( test_bit( ucEventFlags, EVENT_CLOCK_TICK ) )
        {
	      display_time();
	      clear_bit( ucEventFlags, EVENT_CLOCK_TICK );
        }

        if ( test_bit( ucEventFlags, EVENT_POWER_UP ) )
 		  clear_bit( ucEventFlags, EVENT_POWER_UP );

        if ( test_bit( ucEventFlags, EVENT_POWER_DOWN ) )
		{
		  // Power is down
		  clear_bit( PORTC, PC5 );			// diagnostic led off
		  LCD_OFF();						// LCD off
		  ucSystemState = STATE_SLEEP;    
		  clear_bit( ucEventFlags, EVENT_POWER_DOWN );
        }
	    break;

      case STATE_PROGRAM:
	    program_clock();
	    ucSystemState = STATE_RUN;
	    break;

      case STATE_SLEEP:
	  	set_sleep_mode( SLEEP_MODE_PWR_SAVE );
		sleep_mode();
		nop();
		TCCR2B = 0b00000101;		// write the TCCR2 register
        while( ASSR & 0x07 );		// wait until TC2 is updated

        if ( test_bit( ucEventFlags, EVENT_POWER_DOWN ) )
 		  clear_bit( ucEventFlags, EVENT_POWER_DOWN );

        if ( test_bit( ucEventFlags, EVENT_POWER_UP ) )
        {
		  // Power is up, return to run mode
		  LCD_init();						// initialize LCD
		  set_bit( PORTC, PC5 );			// diagnostic led on
		  ucSystemState = STATE_RUN;
		  clear_bit( ucEventFlags, EVENT_POWER_UP );
		}		  
	    break;
	}
  }
}

/*-----------------------------------------------------------------
  system_init()

  PORTB: PB0 (output) LCD D/C
         PB1 (output) LCD reset
         PB2 (output) LCD chip enable
         PB3 (output) ISP MOSI/LCD data in
         PB4 (input) ISP MISO
         PB5 (output) ISP SCK/LCD clock
         PB6 (input) TOSC1 32.768 KHz
         PB7 (input) TOSC2 32.768 KHz

  PORTC: PC0 (input) Encoder "A"
         PC1 (input) Encoder "B"
         PC2 (input) Encoder pushbutton
         PC3 (input) not used
         PC4 (input) not used
         PC5 (output) diagnostic LED
         PC6 (input) ISP !Reset

  PORTD: PD0 (input) not used
         PD1 (input) not used
         PD2 (input) (INT0) Vcc Good
         PD3 (input) not used
         PD4 (input) not used
         PD5 (input) not used
         PD6 (output) Timer/counter2 test output
         PD7 (output) LCD backlight
-----------------------------------------------------------------*/
static void system_init()
{
  int8_t	ii;
  
  // Make sure the clock prescaler is set to 2.
  CLKPR = 0b10000000;       // set clock prescaler change enable
  CLKPR = 0b00000001;       // set prescaler to 2
  
  // Delay 1 second to let clock crystal time to stabilize
  for ( ii=0; ii<15; ++ii )
    _delay_loop_2( 0 );		// 65536 x 250ns x 4 = 65.5ms
  
  // Set up the I/O ports directions, pull-ups
  DDRB  = 0b00101111;		// port B directions
  PORTB = 0b00010000;       // port B initialization, pullups

  DDRC  = 0b00100000;		// port C directions
  PORTC = 0b00011111;       // port C initialization, pullups

  DDRD  = 0b11000000;		// port D directions
  PORTD = 0b00111011;       // port D initialization, pullups
  EICRA = 0b00000001;		// INT0 level change generates interrupt
  EIMSK = 0b00000001;		// enable INT0 interrupt

  // Initialize timer0 for button debouncing
  TCNT0 = INIT_CTR0;
  TCCR0A = 0b00000000;       // timer/counter normal mode
  TCCR0B = 0b00000011;       // prescaler = 64 (turns timer0 on)
  set_bit( TIMSK0, TOIE0 );	 // enable timer/counter0 overflow interrupt

  // Initialize timer/counter 2 as 32KHz oscillator
  ASSR   = 0b00100000;		// use crystal oscillator on TOSC1/2
  TCCR2A = 0;				// normal operation
  TCCR2B = 0b00000101;		// prescaler = 128; start timer
  TCNT2  = 0;				// clear Timer/Counter Register
  while( ASSR & 0x07 );		// wait until TC2 is updated
  TIMSK2 = 0b00000001;		// Timer/Counter2 Interrupt Mask Register

  PRR = 0b10001011;			// power save: no TWI, Timer 1, USART1, USART0, ADC
  sei();					// enable global interrupts
}

/*-----------------------------------------------------------------
  display_time(...)
-----------------------------------------------------------------*/
static void display_time()
{
  static volatile unsigned char ucLastHr=0;
  char			cWork[8], cTime[12];
  unsigned char ucHr, ucX;

  ucHr = ucHours;
  if ( ucFormat == FORMAT_12HRS )
  {
    // Since the clock is in 24 hour format, make adjustment
	if ( !ucHours )
	  ucHr = 12;
	else
	  if ( ucHours > 12 )
	    ucHr = ucHours - 12;
  }

  // Shift right 1 position when there is only one hour digit
  // to show. 
  ucX = ( ucHr < 10 ) ? 2 : 1;

  sprintf( cTime, "%d", (int) ucHr );

  sprintf( cWork, "%02d", (int) ucMinutes );
  strcat( cTime, "." );
  strcat( cTime, cWork );

  sprintf( cWork, "%02d", (int) ucSeconds );
  strcat( cTime, "." );
  strcat( cTime, cWork );

  if ( ucLastHr != ucHr )
  {
	// change of hour can cause shifting
	LCD_clear();
	ucLastHr = ucHr;
  }
  
  LCD_write_string_big( ucX, 1, cTime, MENU_NORMAL );
}


